// SPDX-License-Identifier: MIT

use anyhow::{anyhow, Context, Error, Result};
use axum::extract::ws::{self, WebSocket};
use futures::{
    sink::SinkExt,
    stream::{SplitSink, SplitStream, StreamExt},
};
use std::io::Cursor;
use std::path::Path;
use std::process::{Output, Stdio};
use tokio::io::{AsyncBufReadExt, AsyncWriteExt, BufReader};
use tokio::process::{ChildStdin, Command};
use tracing::{info, trace};

#[cfg(debug_assertions)]
const SCRIPTS_DIR: &str = "./scripts";

#[cfg(not(debug_assertions))]
const SCRIPTS_DIR: &str = "/usr/libexec/abos-web";

fn doas() -> &'static str {
    if cfg!(debug_assertions) && Path::new("./utils/ssh_run.sh").is_file() {
        "./utils/ssh_run.sh"
    } else if Path::new("/usr/bin/doas").is_file() {
        "doas"
    } else {
        ""
    }
}

// take script_dir / doas as argument for tests
fn _command<S>(args: &[S], script_dir: Option<&str>, doas: &str) -> Result<Command>
where
    S: AsRef<std::ffi::OsStr>,
{
    if args.is_empty() {
        return Err(anyhow!("no arg"));
    }

    let mut alloc = String::new();
    let arg0 = match script_dir {
        None => args[0].as_ref(),
        Some(dir) => {
            alloc.push_str(dir);
            alloc.push('/');
            alloc.push_str(
                args[0]
                    .as_ref()
                    .to_str()
                    .context("Could not cast OsStr back to str")?,
            );
            alloc.as_ref()
        }
    };

    if doas.is_empty() {
        let mut command = Command::new(arg0);
        command.args(&args[1..]);
        Ok(command)
    } else {
        let mut command = Command::new(doas);
        command.arg(arg0);
        command.args(&args[1..]);
        Ok(command)
    }
}

pub struct CommandScript(pub bool);
pub struct CommandAsRoot(pub bool);

pub fn command<S>(args: &[S], script: CommandScript, as_root: CommandAsRoot) -> Result<Command>
where
    S: AsRef<std::ffi::OsStr>,
{
    let scripts_dir = if script.0 { Some(SCRIPTS_DIR) } else { None };
    let doas = if as_root.0 { doas() } else { "" };

    _command(args, scripts_dir, doas)
}

/// get next chunk from input if input is set.
/// This is necessary to allow having an optional input in select!,
/// even if we never call it with a none input.
async fn next_if_some(input: &mut Option<&mut SplitStream<&mut WebSocket>>) -> Option<Vec<u8>> {
    match input {
        None => None,
        Some(ref mut input) => match input.next().await {
            None => {
                info!("Websocket closed");
                None
            }
            Some(Err(e)) => {
                info!("Error reading from websocket: {:?}", e);
                None
            }
            Some(Ok(ws::Message::Binary(chunk))) => Some(chunk),
            Some(Ok(ws::Message::Text(text))) if text.as_str() == "END_OF_FILE" => {
                trace!("End of file indicator");
                None
            }
            Some(Ok(msg)) => {
                info!("Unexpected message after file: {:?}", msg);
                None
            }
        },
    }
}

/// likewise: write optional data to optional process stdin.
/// Returns true if we continue or false on error
async fn write_if_some(stdin: &mut Option<ChildStdin>, data: &mut Option<Cursor<Vec<u8>>>) -> bool {
    // stdin and data are guaranteed to be some if we are here
    if let Err(e) = stdin
        .as_mut()
        .unwrap()
        .write_all_buf(&mut data.as_mut().unwrap())
        .await
    {
        info!("Failed to write chunk to process {:?}", e);
        return false;
    };
    true
}

/// Run given command and feed  stdout/stderr to output.
/// If input is not none, feed it to command.
/// Command runs to completion even if output closes early,
/// and this returns exit code.
pub async fn stream_command<S>(
    args: &[S],
    mut input: Option<&mut SplitStream<&mut WebSocket>>,
    output: &mut SplitSink<&mut WebSocket, ws::Message>,
) -> Result<bool>
where
    S: AsRef<std::ffi::OsStr> + std::fmt::Debug,
{
    trace!("Executing (stream) {:?}", args);
    let mut command = command(args, CommandScript(true), CommandAsRoot(true))?
        .stdin(if input.is_some() {
            Stdio::piped()
        } else {
            Stdio::null()
        })
        .stdout(Stdio::piped())
        .stderr(Stdio::piped())
        .spawn()
        .context("Could not start command")?;

    let mut stdin = if input.is_some() {
        Some(command.stdin.take().context("command has no stdin")?)
    } else {
        None
    };
    let mut stdin_data = None;

    let mut stdout = BufReader::new(command.stdout.take().context("command has no stdout")?);
    let mut stdout_buf = vec![];
    let mut stdout_open = true;

    let mut stderr = BufReader::new(command.stderr.take().context("command has no stderr")?);
    let mut stderr_buf = vec![];
    let mut stderr_open = true;

    while stdout_open || stderr_open {
        tokio::select! {
            n = stdout.read_until(b'\n', &mut stdout_buf), if stdout_open => {
                if n.unwrap_or(0) == 0 && stdout_buf.is_empty() {
                    stdout_open = false;
                } else {
                    let text = String::from_utf8_lossy(&stdout_buf);
                    trace!("{}", text);
                    // ignore errors: we continue update anyway
                    let _ = output.send(ws::Message::Text(text.to_string())).await;
                    stdout_buf = vec![];
                }
            },
            n = stderr.read_until(b'\n', &mut stderr_buf), if stderr_open => {
                if n.unwrap_or(0) == 0 && stderr_buf.is_empty() {
                    stderr_open = false;
                } else {
                    let text = String::from_utf8_lossy(&stderr_buf);
                    trace!("{}", text);
                    let _ = output.send(ws::Message::Text(text.to_string())).await;
                    stderr_buf = vec![];
                }
            },
            chunk = next_if_some(&mut input), if stdin.is_some() && stdin_data.is_none() => {
                match chunk {
                    None => {
                        // drop stdin to close it, so process can exit
                        stdin = None;
                    }
                    Some(data) => {
                        stdin_data = Some(Cursor::new(data));
                    },
                }
            },
            cont = write_if_some(&mut stdin, &mut stdin_data), if stdin_data.is_some() => {
                if cont {
                    stdin_data = None
                } else {
                    stdin = None;
                    stdin_data = None
                }
            }
        }
    }

    let status = command.wait().await.context("waiting failed")?;
    Ok(status.success())
}

pub async fn exec_command<S>(args: &[S]) -> Result<Output>
where
    S: AsRef<std::ffi::OsStr> + std::fmt::Debug,
{
    trace!("Executing {:?}", args);
    let output = command(args, CommandScript(true), CommandAsRoot(true))?
        .output()
        .await?;
    if output.status.success() {
        Ok(output)
    } else {
        Err(Error::msg(format!(
            "Command {:?} failed with {}:\nstderr:\n{}\nstdout:\n{}",
            args,
            output.status,
            String::from_utf8_lossy(&output.stderr),
            String::from_utf8_lossy(&output.stdout)
        )))
    }
}

#[cfg(test)]
mod tests {
    use crate::common::process::_command;
    use anyhow::Result;
    use std::process::Command;

    fn check_command(cmd: &Command, arg0: &str, args: &[&str]) {
        assert_eq!(cmd.get_program(), arg0);
        assert_eq!(cmd.get_args().count(), args.len());
        cmd.get_args()
            .zip(args)
            .for_each(|(cmd_arg, check_arg)| assert_eq!(&cmd_arg.to_str().unwrap(), check_arg));
    }

    #[test]
    fn test_command() -> Result<()> {
        assert!(_command(&Vec::<String>::new(), None, "").is_err());

        // single argument
        let args = &["ls"];
        let cmd = _command(args, None, "")?;
        check_command(cmd.as_std(), "ls", &[]);
        let cmd = _command(&args.map(|s| s.to_string()), None, "")?;
        check_command(cmd.as_std(), "ls", &[]);

        let cmd = _command(args, None, "doas")?;
        check_command(cmd.as_std(), "doas", &["ls"]);
        let cmd = _command(&args.map(|s| s.to_string()), None, "doas")?;
        check_command(cmd.as_std(), "doas", &["ls"]);

        let cmd = _command(args, Some("./scripts"), "")?;
        check_command(cmd.as_std(), "./scripts/ls", &[]);
        let cmd = _command(&args.map(|s| s.to_string()), Some("./scripts"), "")?;
        check_command(cmd.as_std(), "./scripts/ls", &[]);

        let cmd = _command(args, Some("./scripts"), "doas")?;
        check_command(cmd.as_std(), "doas", &["./scripts/ls"]);
        let cmd = _command(&args.map(|s| s.to_string()), Some("./scripts"), "doas")?;
        check_command(cmd.as_std(), "doas", &["./scripts/ls"]);

        // some argument
        let args = &["ls", "-l"];
        let cmd = _command(args, None, "")?;
        check_command(cmd.as_std(), "ls", &["-l"]);
        let cmd = _command(&args.map(|s| s.to_string()), None, "")?;
        check_command(cmd.as_std(), "ls", &["-l"]);

        let cmd = _command(args, None, "doas")?;
        check_command(cmd.as_std(), "doas", &["ls", "-l"]);
        let cmd = _command(&args.map(|s| s.to_string()), None, "doas")?;
        check_command(cmd.as_std(), "doas", &["ls", "-l"]);

        let cmd = _command(args, Some("./scripts"), "")?;
        check_command(cmd.as_std(), "./scripts/ls", &["-l"]);
        let cmd = _command(&args.map(|s| s.to_string()), Some("./scripts"), "")?;
        check_command(cmd.as_std(), "./scripts/ls", &["-l"]);

        let cmd = _command(args, Some("./scripts"), "doas")?;
        check_command(cmd.as_std(), "doas", &["./scripts/ls", "-l"]);
        let cmd = _command(&args.map(|s| s.to_string()), Some("./scripts"), "doas")?;
        check_command(cmd.as_std(), "doas", &["./scripts/ls", "-l"]);

        Ok(())
    }
}
