// SPDX-License-Identifier: MIT

use askama::Template;
use axum::{
    extract::ws::{self, WebSocket, WebSocketUpgrade},
    response::{IntoResponse, Redirect},
    routing::get,
    Router,
};
use axum_sessions::extractors::ReadableSession;
use futures::{sink::SinkExt, stream::StreamExt};
use tracing::{trace, warn};

use crate::common::{exec_command, html_template::HtmlTemplate, stream_command, Config};
use crate::error::{ErrorString, PageResult};

mod version;
pub use version::{get_last_update, get_swu_versions, SwuLastUpdate, SwuVersion};

#[derive(Template)]
#[template(path = "../src/swu/templates/swu.html")]
struct SwuTemplate {
    initial_setup_installed: bool,
    swupdate_watch_url: String,
    swu_versions: Vec<SwuVersion>,
    swu_last_update: Option<SwuLastUpdate>,
}

pub trait SwuRoutes {
    fn swu_routes(self) -> Self;
}

impl SwuRoutes for Router {
    fn swu_routes(self) -> Self {
        self.route("/swu", get(swu)).route("/swu_ws", get(swu_ws))
    }
}

async fn swu(session: ReadableSession) -> PageResult {
    if !cfg!(debug_assertions) && !session.get::<bool>("login").unwrap_or(false) {
        return Ok(Redirect::to("/").into_response());
    }

    let swu_versions = get_swu_versions().await.unwrap_or_default();
    let swu_last_update = get_last_update(&swu_versions).await;
    // note: cannot use await after getting this lock
    let swu_config = &Config::get().read()?.swu;
    let template = SwuTemplate {
        initial_setup_installed: swu_config.initial_setup_installed,
        swupdate_watch_url: swu_config.watch_urls.join(" "),
        swu_versions,
        swu_last_update,
    };
    Ok(HtmlTemplate::new("SWUインストール", template).into_response())
}

async fn swu_ws(session: ReadableSession, ws: WebSocketUpgrade) -> impl IntoResponse {
    if !cfg!(debug_assertions) && !session.get::<bool>("login").unwrap_or(false) {
        // close socket immediately
        return Err(ErrorString {
            message: "not logged in".to_string(),
        });
    }
    Ok(ws.on_upgrade(swu_socket_handler))
}

async fn swu_socket_handler(mut socket: WebSocket) {
    // websocket "protocol"
    // first message should be Text() with a keyword:
    //  * "file": second message onwards are binary file content
    //  * "url": second message onwards are url (space separated), terminated by "END"
    //  * "log": no more message accepted
    let Some(Ok(ws::Message::Text(command))) = socket.recv().await else {
        warn!("Socket first message was not text");
        return;
    };
    trace!("Got {} on websocket", command);
    match command.as_str() {
        "file" => swu_file(&mut socket).await,
        "url" => swu_url(&mut socket).await,
        "latest_log" => swu_latest_log(&mut socket).await,
        _ => warn!("Invalid command on websocket: {}", command),
    }
    // wait for close from client: if we close first,
    // the browser will not be able to process all the messages until the end
    let _ = socket.send(ws::Message::Close(None)).await;
    while let Some(Ok(_)) = socket.recv().await {}
}

async fn swu_file(socket: &mut WebSocket) {
    let (mut output, mut input) = socket.split();
    if let Err(e) = stream_command(&["swupdate.sh", "stdin"], Some(&mut input), &mut output).await {
        warn!("Error streaming swupdate file: {:?}", e);
        return;
    };
    let _ = output
        .send(ws::Message::Text("swupdate exited".to_string()))
        .await;
}

async fn swu_url(socket: &mut WebSocket) {
    let (mut output, mut input) = socket.split();
    let mut cmd = vec!["swupdate.sh".to_string(), "url".to_string()];
    'outer: while let Some(Ok(ws::Message::Text(url))) = input.next().await {
        for u in url.trim().split(' ') {
            match u {
                "END" => break 'outer,
                _ => cmd.push(u.to_string()),
            }
        }
    }
    if cmd.len() == 2 {
        let _ = output
            .send(ws::Message::Text("No URL given?".to_string()))
            .await;
        return;
    }
    trace!("Starting {:?}", cmd);
    if let Err(e) = stream_command(&cmd, None, &mut output).await {
        warn!("Error streaming swupdate url: {:?}", e);
        return;
    };
    let _ = output
        .send(ws::Message::Text("swupdate exited".to_string()))
        .await;
}
async fn swu_latest_log(socket: &mut WebSocket) {
    let log = match exec_command(&["get_latest_swupdate_log.sh"]).await {
        Ok(status) => String::from_utf8_lossy(&status.stdout).to_string(),
        Err(_) => "ログを取得できませんでした".to_string(),
    };
    let _ = socket.send(ws::Message::Text(log)).await;
}
